import os

from habutax.enum import filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form1040S3(Form):
    form_name = "1040_s3"
    tax_year = 2022
    description = "Schedule 3 (Form 1040)"
    long_description = "Additional Credits and Payments"
    jurisdiction = Jurisdiction.US
    sequence_no = 3

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('other_foreign_gross_income', description="Do you have any foreign source gross income *not* from interest and dividends or any foreign tax paid on that income not reported to you on Form 1099-INT, Form 1099-DIV?"),
            BooleanInput('child_dependent_care', description="Did you pay expenses for the care of a qualifying individual (child or other dependent) to enable you (and your spouse, if filing a joint return) to work or actively look for work?"),
            BooleanInput('education_credit', description="Did you (or your dependent) pay qualified expenses in 2022 for yourself, your spouse, or your dependent to enroll in or attend an eligible educational institution? If so, do you need to claim an education credit on Form 8863?"),
            BooleanInput('retirement_savings_contributions', description="Did you, or your spouse if filing jointly, make (a) contributions, other than rollover contributions, to a traditional or Roth IRA; (b) elective deferrals to a 401(k) or 403(b) plan (including designated Roth contributions) or to a governmental 457, SEP, or SIMPLE plan; (c) voluntary employee contributions to a qualified retirement plan (including the federal Thrift Savings Plan); (d) contributions to a 501(c)(18)(D) plan; or (e) contributions to an ABLE account by the designated beneficiary, as defined in section 529A?"),
            BooleanInput('residential_energy_credit', description="Did you pay for any energy-efficient upgrades to your home in 2022? This might be things like installing solar panels, wind energy installations, adding insulation or high-efficiency HVAC units."),
            BooleanInput('other_nonrefundable_credits', description="Do you have any less common nonrefundable credits to claim? These include general business credits (Form 3800), prior year minimum tax (Form 8801), adoption credit (Form 8839), credit for the elderly or disabled (Schedule R), alternative motor vehicle credit (Form 8910), qualified plug-in motor vehicle credit (Form 8936), mortgage interest credit (Form 8396), District of Columbia first-time homebuyer credit (Form 8859), qualified electric vehicle credit (Form 8834), alternative fuel vehicle refueling property credit (Form 8911), credit to holders of tax credit bonds (Form 8912), amount on Form 8978, line 14 (reporting for partnership income), or any other nonrefundable tax credits."),
        ]

        def line_1(self, i, v):
            foreign_tax = float(sum([v[f'1099-int:{n}.box_6'] for n in range(i['1040.number_1099-int'])]))
            foreign_tax += float(sum([v[f'1099-div:{n}.box_7'] for n in range(i['1040.number_1099-div'])]))
            f1116_limit = 600.0 if i['1040.filing_status'] is filing_status.MarriedFilingJointly else 300.0
            if i['other_foreign_gross_income'] or foreign_tax > f1116_limit:
                return self.not_implemented()
            return foreign_tax if foreign_tax > 0.001 else None

        def retirement_savings_limit(self, i, v):
            if i['1040.filing_status'] is filing_status.MarriedFilingJointly:
                return 68000.0
            elif i['1040.filing_status'] is filing_status.HeadOfHousehold:
                return 51000.0
            else:
                return 34000.0

        optional_fields = [
            FloatField('1', line_1),
            FloatField('2', lambda s, i, v: s.not_implemented() if i['child_dependent_care'] else None), # Form 2441, line 11
            FloatField('3', lambda s, i, v: s.not_implemented() if i['education_credit'] else None), # Form 8863, line 19
            FloatField('4', lambda s, i, v: s.not_implemented() if v['1040.11'] <= retirement_savings_limit(s, i, v) and i['retirement_savings_contributions'] else None), # Form 8880
            FloatField('5', lambda s, i, v: s.not_implemented() if i['residential_energy_credit'] else None), # Form 5695
            FloatField('6a', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 3800
            FloatField('6b', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8801
            FloatField('6c', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8839
            FloatField('6d', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form Schedule R
            FloatField('6e', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8901
            FloatField('6f', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8936
            FloatField('6g', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8396
            FloatField('6h', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8859
            FloatField('6i', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8834
            FloatField('6j', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8911
            FloatField('6k', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8912
            FloatField('6l', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Form 8978, line 14
            FloatField('6z', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Other nonrefundable credits
            FloatField('7', lambda s, i, v: sum([v[f'6{l}'] for l in "abcdefghijklz"])),
            FloatField('8', lambda s, i, v: sum([v[f'{n}'] for n in range(1,5+1)]) + v['7']),
        ]
        required_fields = [
            StringField('6z_type', lambda s, i, v: s.not_implemented() if i['other_nonrefundable_credits'] else None), # Type of any nonrefundable credits
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_01[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_02[0]', '1040.you_ssn', max_length=11),
            TextPDFField('topmostSubform[0].Page1[0].f1_03[0]', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_04[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_05[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_06[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].f1_07[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].f1_08[0]', '6a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_09[0]', '6b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '6c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_11[0]', '6d'),
            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '6e'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '6f'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '6g'),
            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '6h'),
            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '6i'),
            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '6j'),
            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '6k'),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '6l'),
            TextPDFField('topmostSubform[0].Page1[0].Line6z_ReadOrder[0].f1_20[0]', '6z_type'),
            TextPDFField('topmostSubform[0].Page1[0].Line6z_ReadOrder[0].f1_21[0]', '6z_type'),
            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '6z'),
            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '8'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_01[0]', '9'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_02[0]', '10'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_03[0]', '11'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_04[0]', '12'),
#            TextPDFField('topmostSubform[0].Page2[0].Line13_ReadOrder[0].f2_05[0]', '13a'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_06[0]', '13b'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_07[0]', '13c'), #"Reserved for future use"
#            TextPDFField('topmostSubform[0].Page2[0].f2_08[0]', '13d'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_09[0]', '13e'), #"Reserved for future use"
#            TextPDFField('topmostSubform[0].Page2[0].f2_10[0]', '13f'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_11[0]', '13g'), #"Reserved for future use"
#            TextPDFField('topmostSubform[0].Page2[0].f2_12[0]', '13h'),
#            TextPDFField('topmostSubform[0].Page2[0].Line13z_ReadOrder[0].f2_14[0]', '13z_type'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_15[0]', '13z'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_16[0]', '14'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_17[0]', '15'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'f1040s3.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



import unittest

from ..test_form import FormTestCase

import habutax.forms.ty2021 as ty2021
from habutax import enum

class FormW2TestCase(FormTestCase):
    def setUp(self):
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/fw_2_0.habutax')

    def test_simple_w2(self):
        solution = self.assertSolve(['w-2:0'])

        self.assertIn('w-2:0', solution)

        self.assertEnumFieldEqual(solution, 'w-2:0.belongs_to', enum.taxpayer_or_spouse, enum.taxpayer_or_spouse.spouse)
        self.assertStringFieldEqual(solution, 'w-2:0.box_c', "Fake County Public School System")
        self.assertStringFieldEqual(solution, 'w-2:0.box_e', "Suzy Llama")
        self.assertDollarsEqual(solution, 'w-2:0.box_1', 12595.47)
        self.assertDollarsEqual(solution, 'w-2:0.box_2', 905.24)
        self.assertDollarsEqual(solution, 'w-2:0.box_3', 17763.74)
        self.assertDollarsEqual(solution, 'w-2:0.box_4', 1101.35)
        self.assertDollarsEqual(solution, 'w-2:0.box_5', 17763.74)
        self.assertDollarsEqual(solution, 'w-2:0.box_6', 257.57)
        self.assertDollarsEqual(solution, 'w-2:0.box_7', 0.00)
        self.assertDollarsEqual(solution, 'w-2:0.box_8', 0.00)
        self.assertDollarsEqual(solution, 'w-2:0.box_10', 0.00)
        self.assertDollarsEqual(solution, 'w-2:0.box_11', 0.00)
        self.assertBooleanFieldEqual(solution, 'w-2:0.box_13_retirement', True)
        self.assertBooleanFieldEqual(solution, 'w-2:0.box_13_sick_day', False)
        self.assertBooleanFieldEqual(solution, 'w-2:0.box_13_statutory', False)
        self.assertEnumFieldEqual(solution, 'w-2:0.box_15', enum.us_states, enum.us_states.SC)
        self.assertDollarsEqual(solution, 'w-2:0.box_16', 12595.47)
        self.assertDollarsEqual(solution, 'w-2:0.box_17', 455.00)
        self.assertDollarsEqual(solution, 'w-2:0.box_18', 0.00)
        self.assertDollarsEqual(solution, 'w-2:0.box_19', 0.00)



import configparser
import unittest

from . import context

import habutax.enum as enum
from habutax.inputs import *
from habutax.form import Form

class InputTestForm(Form):
    form_name = "0000"
    tax_year = 1970

    def __init__(self, **kwargs):
        super().__init__(__class__, [], [], [], **kwargs)

class InputTestCase(unittest.TestCase):
    def setUp(self):
        self.config = configparser.ConfigParser()
        self.form = InputTestForm()

    def tearDown(self):
        del self.config
        del self.form

    def test_string_input(self):
        str_input = StringInput('name', description="What is your name?")
        str_input.__form_init__(self.form)

        for value in ['John Smith ', ' ', '839']:
            self.assertTrue(str_input.valid(value))
            self.assertIsInstance(str_input.value(value), str)
            self.assertEqual(str_input.value(value), value.strip())

    def test_boolean_input(self):
        bool_input = BooleanInput('snake', description="Are you a snake?")
        bool_input.__form_init__(self.form)

        for invalid in ['nope', 'nah', 'maybe', 'tomorrow', 'definitely', '', '2']:
            self.assertFalse(bool_input.valid(invalid))

        for yes in ['yes', 'Yes', 'true', 'True', 'TRUE', '1', 'ON', 'on']:
            self.assertTrue(bool_input.valid(yes))
            self.assertIsInstance(bool_input.value(yes), bool)
            self.assertEqual(bool_input.value(yes), True)

        for no in ['no', 'NO', 'false', 'FALSE', 'False', '0', 'Off', 'OFF']:
            self.assertTrue(bool_input.valid(no))
            self.assertIsInstance(bool_input.value(no), bool)
            self.assertEqual(bool_input.value(no), False)

    def test_integer_input(self):
        int_input = IntegerInput('snakes', description="How many snakes are on your plane?")
        int_input.__form_init__(self.form)

        for invalid in ['not on my plane', '-', '+', '$', '0.4']:
            self.assertFalse(int_input.valid(invalid))

        for valid in ['-1', '3', '    51', '92838927359872']:
            self.assertTrue(int_input.valid(valid))
            self.assertIsInstance(int_input.value(valid), int)
            self.assertEqual(int_input.value(valid), int(valid.strip()))

        self.assertTrue(int_input.valid(''))
        self.assertIsInstance(int_input.value(''), int)
        self.assertEqual(int_input.value(''), 0)

    def test_float_input(self):
        float_input = FloatInput('fractional_snakes', description="How many snakes are on your plane?")
        float_input.__form_init__(self.form)

        for invalid in ['not on my plane', '-', '+', '$', '9 4', '84/']:
            self.assertFalse(float_input.valid(invalid))

        for valid in ['-0.14', '9\t', '    51', '92838927359872', '0.83858']:
            self.assertTrue(float_input.valid(valid))
            self.assertIsInstance(float_input.value(valid), float)
            self.assertAlmostEqual(float_input.value(valid), float(valid.strip()), places=5)

        self.assertTrue(float_input.valid(''))
        self.assertIsInstance(float_input.value(''), float)
        self.assertEqual(float_input.value(''), 0.0)

    def test_enum_input(self):
        colors = enum.make('Colors of the rainbow', {
            'Red': 'the color you bleed',
            'Orange': 'tasty fruit',
            'Yellow': 'like the sun',
            'Green': 'leaves in the summer',
            'Blue': 'cool ocean breeze',
            'Indigo': 'purple',
            'Violet': 'also purple',
        })
        enum_color = EnumInput('favorite_color', colors, description="What is your favorite color?")
        enum_color.__form_init__(self.form)

        for invalid in ['', 'green', 'GREEN', 'sky blue']:
            self.assertFalse(enum_color.valid(invalid))

        for valid, result in {'Red': colors.Red, 'Green': colors.Green, 'Blue': colors.Blue}.items():
            self.assertTrue(enum_color.valid(valid))
            self.assertIsInstance(enum_color.value(valid), enum.StringyEnum)
            self.assertEqual(enum_color.value(valid), result)

    def test_regex_input(self):
        regex_name = RegexInput('name', '^(Bob|[Rr]obert) Smith$', description="What is your name?")
        regex_name.__form_init__(self.form)

        for invalid in ['Bobert Smith', 'Bob', 'me']:
            self.assertFalse(regex_name.valid(invalid))

        for valid in ['Bob Smith', 'Robert Smith', 'robert Smith']:
            self.assertTrue(regex_name.valid(valid))
            self.assertIsInstance(regex_name.value(valid), str)
            self.assertEqual(regex_name.value(valid), valid)

        regex_routing = RegexInput('routing_number', '^(0[1-9]|1[0-2]|2[1-9]|3[0-2])[0-9]{7}$', description="What is your routing number?")
        regex_routing.__form_init__(self.form)

        for invalid in ['00', '000000000', '158374958']:
            self.assertFalse(regex_routing.valid(invalid))

        for valid in ['012345678', '310000000', '258377691']:
            self.assertTrue(regex_routing.valid(valid))
            self.assertEqual(regex_routing.value(valid), str(valid))

    def test_ssn_input(self):
        ssn = SSNInput('taxpayer_ssn', description="Your social security number")
        ssn.__form_init__(self.form)

        for invalid in ['1234567890', '12345678', 'hello?']:
            self.assertFalse(ssn.valid(invalid))

        for valid, result in {'000-00-0000': '000000000', '123456789': '123456789'}.items():
            self.assertTrue(ssn.valid(valid))
            self.assertEqual(ssn.value(valid), result)



import habutax.enum as enum
from habutax.form import InputForm, Jurisdiction
from habutax.inputs import *

class Form1098(InputForm):
    form_name = "1098"
    tax_year = 2022
    description = "Form 1098"
    long_description = "Mortgage Interest Statement"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            EnumInput('belongs_to', enum.taxpayer_spouse_or_both, description="To whom does this 1098 belong?"),
            StringInput('recipient_lender_name', description="Recipient/Lender's name: To whom did you pay this mortgage interest?"),
            StringInput('payer_borrower', description="Payer/Borrower: Name of borrower who paid the interest"),
            FloatInput('box_1', description="Mortgage interest received from payer(s)/borrower(s)"),
            FloatInput('box_2', description="Outstanding mortgage principal"),
            StringInput('box_3', description="Mortgage origination date"),
            FloatInput('box_4', description="Refund of overpaid interest"),
            FloatInput('box_5', description="Mortgage insurance premiums"),
            FloatInput('box_6', description="Points paid on purchase of principal residence"),
            BooleanInput('box_7', description="Is the checkbox on box 7 checked?"),
            StringInput('box_8', description="Address or description of property securing mortgage"),
            IntegerInput('box_9', description="Number of properties securing the mortgage"),
            FloatInput('box_10', description="Other"),
            StringInput('box_11', description="Mortgage acquisition date"),
        ]

        super().__init__(__class__, inputs, **kwargs)



import unittest

from .. import context

from habutax.forms.ty2022.f1040_figure_tax import figure_tax
from habutax import enum

class FigureTaxTestCase(unittest.TestCase):
    def test_figure_2022_1040_tax(self):
        scenarios = (
            (75047.08, enum.filing_status.Single, 12123),
            (98325.84, enum.filing_status.MarriedFilingJointly, 12866),
            (99999.99, enum.filing_status.HeadOfHousehold, 16330.00),
            (100000.00, enum.filing_status.HeadOfHousehold, 16336.00),
            (588939, enum.filing_status.QualifyingSurvivingSpouse, 153634.65),
            (588939, enum.filing_status.MarriedFilingJointly, 153634.65),
            (1482.40, enum.filing_status.MarriedFilingSeparately, 149),
            (8182983.94, enum.filing_status.MarriedFilingSeparately, 2994978.56),
        )
        for taxable, status, expected_tax in scenarios:
            tax = figure_tax(taxable, status)
            self.assertAlmostEqual(expected_tax, tax, places=2)

    def test_figure_2022_1040_tax_monotonically_increasing(self):
        statuses = [enum.filing_status.Single, enum.filing_status.MarriedFilingSeparately, enum.filing_status.MarriedFilingJointly, enum.filing_status.QualifyingSurvivingSpouse, enum.filing_status.HeadOfHousehold]
        for status in statuses:
            last = 0
            for taxable_income in range(0, 200, 3):
                this = figure_tax(taxable_income, status)
                self.assertGreaterEqual(this, last)
                last = this
            for taxable_income in range(200, 10000, 23):
                this = figure_tax(taxable_income, status)
                self.assertGreaterEqual(this, last)
                last = this
            for taxable_income in range(10000, 199999, 43):
                this = figure_tax(taxable_income, status)
                self.assertGreaterEqual(this, last)
                last = this
            for taxable_income in range(199999, 750000, 1967):
                this = figure_tax(taxable_income, status)
                self.assertGreaterEqual(this, last)
                last = this



import os

from habutax.enum import filing_status_2021 as filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form1040SA(Form):
    form_name = "1040_sa"
    tax_year = 2021
    description = "Schedule A (Form 1040)"
    long_description = "Itemized Deductions"
    jurisdiction = Jurisdiction.US
    sequence_no = 7

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('itemize_though_less', description="Do you want to itemize deductions even though they are less than your standard deduction?"),
            FloatInput('medical_dental_expenses', description="Enter the total amount of medical and dental expenses you are allowed to deduct for 2021. This can include insurance premiums for medical and dental care, prescription medicines or insulin, medical specialists, examinations, or other medical care. See the instructions for Form 1040, Schedule A for a detailed explanation of the limitations and restrictions"),
            BooleanInput('general_sales_tax', description="Do you want to deduct general sales tax instead of state and local income tax? (You must choose one, and choosing to deduct general sales tax means your return is not supported by Habutax)"),
            FloatInput('state_local_real_estate_taxes', description="Enter the sum of any state and local taxes you paid in 2021 (either directly or through your mortgage company) on real estate you own that wasn't used for business. See instructions for Form 1040, Schedule A, line 5b for details of what can and cannot be claimed."),
            FloatInput('state_local_personal_property_taxes', description="Enter the sum of any state and local personal property taxes you paid in 2021, but only if the taxes were based on value alone and were imposed on a yearly basis. If you paid a yearly fee for the registration of your car and part of that fee was based on the car's value and part was based on its weight, you can deduct only the part of the fee that was based on the car's value."),
            FloatInput('other_taxes_amount', description="Enter the total amount of any 'other taxes' you want to deduct. You may include income taxes you paid to a foreign country and generation skipping tax (GST) imposed on certain income distributions, but you may want to take a credit for the foreign tax instead of a deduction. See the instructions for Schedule 3 (Form 1040), line 1, for details."),
            StringInput('other_taxes_type', description="Enter the 'type' of each 'other taxes' you are deducting."),
            BooleanInput('loan_limitations', description="Do you have mortgage loans taken out after October 13, 1987, that exceed $750,000 ($375,000 if you are married filing separately, loans exceeding the fair market value of the home, or loan proceeds not used to buy, build, or substantially improve your home?"),
            FloatInput('other_mortgage_interest', description="Enter any home mortgage interest not reported to you on Form 1098."),
            StringInput('other_mortgage_interest_desc', description="If the home interest you are claiming that was not reported to you on 1098 was paid to the person from whom you bought the home, see instructions and enter that person’s name, identifying no., and address here"),
            FloatInput('other_mortgage_points', description="Enter any home mortgage points not reported to you on Form 1098. Limits may apply, see Pub. 936 for details."),
            BooleanInput('mortgage_insurance_premiums_special', description="Did you share the costs for mortgage insurance premiums reported on form 1098 with someone (other than your spouse if married filing jointly), or pay mortgage insurance premiums in 2021 for a year other than 2021?"),
            BooleanInput('investment_interest', description="Did you pay interest on money you borrowed that is allocable to property held for investment in 2021?"),
            FloatInput('charitable_cash_check', description="Enter the total amount of all charitable gifts to Qualified Charitable Organizations you made in 2021 via cash or check. Ensure you have (and save for your records) contempotaneous written acknowledgement of any individual gift over $250. See the instructions for line 11 of Form 1040, Schedule A if you have any questions."),
            FloatInput('charitable_other_than_cash_check', description="Enter the total amount of all charitable gifts to Qualified Charitable Organizations you made in 2021 via methods *other than* cash or check. Ensure you have (and save for your records) contempotaneous written acknowledgement of any individual gift over $250. See the instructions for line 12 of Form 1040, Schedule A and/or Pub. 526 if you have any questions."),
            BooleanInput('filling_8283', description="You indicated you gave total noncash charitable gifts of over $500 in 2021. You must file one or more Forms 8283 if the amount of your deduction for each noncash contribution is more than $500. You must also file Form 8283 if you have a group of similar items for which a total deduction of over $500 is claimed. Do you agree to fill out Form(s) 8283 as appropriate and include them with your tax return (you will *not* be reminded again)? See the instructions for line 12 of Form 1040, Schedule A and Pub. 526."),
            FloatInput('charitable_carryover', description="Enter any carryover for any charitable giving you were not allowed to deduct in the previous 5 years because it exceeded the amount you were allowed to decuct. See the instructions for line 12 of Form 1040, Schedule A and Pub. 526."),
            BooleanInput('casualty_theft', description="Do you have any casualty and theft loss(es) from a federally declared disaster (other than net qualified disaster losses) to report?"),
            FloatInput('other_itemized', description="Enter the amount of any other itemized deductions you want to claim. See instructions for Form 1040, Schedule A, line 13"),
            StringInput('other_itemized_type', description="Enter the type of any other itemized deductions you are claiming. See instructions for Form 1040, Schedule A, line 13"),
        ]

        def line_5a(self, i, v):
            # TODO If we implement Schedules C, E, or F, we must reduce the
            # amount calculated here by any amounts of these withheld taxes
            # which are deducted on one of those forms (or elsewhere)
            state_local_income_taxes = sum(v[f'w-2:{n}.box_17'] + v[f'w-2:{n}.box_19'] for n in range(i['1040.number_w-2']))
            state_local_income_taxes += sum(v[f'1099-div:{n}.box_16_1'] + v[f'1099-div:{n}.box_16_2'] for n in range(i['1040.number_1099-div']))
            state_local_income_taxes += sum(v[f'1099-int:{n}.box_17_1'] + v[f'1099-int:{n}.box_17_2'] for n in range(i['1040.number_1099-int']))
            state_local_income_taxes += sum(v[f'1099-r:{n}.box_14_1'] + v[f'1099-r:{n}.box_14_2'] + v[f'1099-r:{n}.box_17_1'] + v[f'1099-r:{n}.box_17_2'] for n in range(i['1040.number_1099-r']))
            return state_local_income_taxes

        def line_8a(self, i, v):
            mortgage_interest_points = sum([v[f'1098:{n}.box_1'] for n in range(i['1040.number_1098'])])
            mortgage_interest_points += sum([v[f'1098:{n}.box_6'] for n in range(i['1040.number_1098'])])
            if mortgage_interest_points > 0.001 and i['loan_limitations']:
                self.not_implemented()
            return mortgage_interest_points

        def line_8d(self, i, v):
            mortgage_insurance_premiums = sum([v[f'1098:{n}.box_5'] for n in range(i['1040.number_1098'])])
            if mortgage_insurance_premiums < 0.001:
                return None
            premium_limit = 50000 if i['1040.filing_status'] == filing_status.MarriedFilingSeparately else 100000
            if mortgage_insurance_premiums > 0.001 and i['mortgage_insurance_premiums_special'] and v['1040:11'] > premium_limit:
                self.not_implemented()
            return mortgage_insurance_premiums

        def line_12(self, i, v):
            if i['charitable_other_than_cash_check'] > 500 and not i['filling_8283']:
                self.not_implemented()
            return i['charitable_other_than_cash_check']

        def line_18(self, i, v):
            penalties = sum([v[f'1099-int:{n}.box_2'] for n in range(i['1040.number_1099-int'])])
            penalties += sum([v[f'1099-oid:{n}.box_3'] for n in range(i['1040.number_1099-oid'])])
            if penalties > 0.01:
                return penalties
            return ""

        optional_fields = [
            # Medical and Dental Expenses
            FloatField('1', lambda s, i, v: i['medical_dental_expenses']),
            FloatField('2', lambda s, i, v: v['1040.11']),
            FloatField('3', lambda s, i, v: 0.075 * v['2']),
            FloatField('4', lambda s, i, v: 0.0 if v['3'] > v['1'] else v['3'] - v['1']),

            # Taxes you paid
            BooleanField('5a_checkbox', lambda s, i, v: False if not i['general_sales_taxes'] else s.not_implemented()),
            FloatField('5a', line_5a),
            FloatField('5b', lambda s, i, v: i['state_local_real_estate_taxes']),
            FloatField('5c', lambda s, i, v: i['state_local_personal_property_taxes']),
            FloatField('5d', lambda s, i, v: v['5a'] + v['5b'] + v['5c']),
            FloatField('5e', lambda s, i, v: min(5000.0 if i['1040.filing_status'] == filing_status.MarriedFilingSeparately else 10000.0, v['5d'])),
            FloatField('6', lambda s, i, v: i['other_taxes_amount']),
            FloatField('7', lambda s, i, v: v['5e'] + v['6']),

            # Interest you paid
            FloatField('8a', line_8a),
            FloatField('8b', lambda s, i, v: i['other_mortgage_interest'] if i['other_mortgage_interest'] > 0.001 and not i['loan_limitations'] else None),
            FloatField('8c', lambda s, i, v: i['other_mortgage_points'] if i['other_mortgage_points'] > 0.001 and not i['loan_limitations'] else None),
            FloatField('8d', line_8d),
            FloatField('8e', lambda s, i, v: v['8a'] + v['8b'] + v['8c'] + v['8d']),
            FloatField('9', lambda s, i, v: s.not_implemented() if i['investment_interest'] else None),
            FloatField('10', lambda s, i, v: v['8e'] + v['9']),

            # Gifts to Charity
            FloatField('11', lambda s, i, v: i['charitable_cash_check']),
            FloatField('12', line_12),
            FloatField('13', lambda s, i, v: i['charitable_carryover'] if i['charitable_carryover'] > 0.001 else None),
            FloatField('14', lambda s, i, v: v['11'] + v['12'] + v['13']),

            # Casualty and Theft Losses
            FloatField('15', lambda s, i, v: s.not_implemented() if i['casualty_theft'] else None),

            # Other Itemized Deductions
            FloatField('16', lambda s, i, v: i['other_itemized'] if i['other_itemized']  > 0.001 else None),

            # Total Itemized Deductions
            FloatField('17', lambda s, i, v: v['4'] + v['7'] + v['10'] + v['14'] + v['15'] + v['16']),
            BooleanField('18', lambda s, i, v: i['itemize_though_less']),
        ]
        required_fields = [
            StringField('6_type', lambda s, i, v: i['other_taxes_type'] if i['other_taxes_amount'] > 0.001 else None),
            StringField('8b_desc', lambda s, i, v: i['other_mortgage_interest_desc'] if i['other_mortgage_interest'] > 0.001 and not i['loan_limitations'] else None),
            StringField('16_type', lambda s, i, v: i['other_itemized_type'] if i['other_itemized'] > 0.001 else None),
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_1[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_2[0]', '1040.you_ssn', max_length=11),
            TextPDFField('topmostSubform[0].Page1[0].f1_3[0]', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_4[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_5[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_6[0]', '4'),
#            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[0]', '5a_checkbox', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_7[0]', '5a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_8[0]', '5b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_9[0]', '5c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '5d'),
            TextPDFField('topmostSubform[0].Page1[0].f1_11[0]', '5e'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '6_type'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '6_type'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '7'),
#            ButtonPDFField('topmostSubform[0].Page1[0].Line8_ReadOrder[0].c1_2[0]', '8', '1'), # If you didn’t use all of your home mortgage loan(s) to buy, build, or improve your home, see instructions and check this box
            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '8a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '8b_desc'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '8b_person'),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '8b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_20[0]', '8c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_21[0]', '8d'),
            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '8e'),
            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '10'),
            TextPDFField('topmostSubform[0].Page1[0].f1_25[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_26[0]', '12'),
            TextPDFField('topmostSubform[0].Page1[0].f1_27[0]', '13'),
            TextPDFField('topmostSubform[0].Page1[0].f1_28[0]', '14'),
            TextPDFField('topmostSubform[0].Page1[0].f1_29[0]', '15'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_30[0]', '16_type'),
            TextPDFField('topmostSubform[0].Page1[0].f1_31[0]', '16_type'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_32[0]', '16_type'),
            TextPDFField('topmostSubform[0].Page1[0].f1_33[0]', '16'),
            TextPDFField('topmostSubform[0].Page1[0].f1_34[0]', '17'),
            ButtonPDFField('topmostSubform[0].Page1[0].Line18_ReadOrder[0].c1_3[0]', '18', '1'),
        ]

        pdf_file = os.path.join(os.path.dirname(__file__), 'f1040sa.pdf')
        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return values['1040.itemizing']



from collections.abc import Mapping
from enum import IntEnum, auto, unique

from habutax.inputs import StringInput, BooleanInput, IntegerInput, FloatInput, EnumInput, SSNInput
from habutax.fields import StringField, BooleanField, IntegerField, FloatField, EnumField

class AutoNumber(IntEnum):
    def _generate_next_value_(name, start, count, last_values):
        return 0 if len(last_values) == 0 else max(last_values) + 1

@unique
class Jurisdiction(AutoNumber):
    # US Federal
    US = auto()
    # US state
    AK = auto()
    AL = auto()
    AR = auto()
    AZ = auto()
    CA = auto()
    CO = auto()
    CT = auto()
    DC = auto()
    DE = auto()
    FL = auto()
    GA = auto()
    HI = auto()
    IA = auto()
    ID = auto()
    IL = auto()
    IN = auto()
    KS = auto()
    KY = auto()
    LA = auto()
    MA = auto()
    MD = auto()
    ME = auto()
    MI = auto()
    MN = auto()
    MO = auto()
    MS = auto()
    MT = auto()
    NC = auto()
    ND = auto()
    NE = auto()
    NH = auto()
    NJ = auto()
    NM = auto()
    NV = auto()
    NY = auto()
    OH = auto()
    OK = auto()
    OR = auto()
    PA = auto()
    PR = auto()
    RI = auto()
    SC = auto()
    SD = auto()
    TN = auto()
    TX = auto()
    UT = auto()
    VA = auto()
    VT = auto()
    WA = auto()
    WI = auto()
    WV = auto()
    WY = auto()

class Form(object):
    def __init__(self,
                 child_cls,
                 inputs,
                 required_fields,
                 optional_fields,
                 pdf_fields=[],
                 pdf_file=None,
                 instance=None,
                 solver=None):

        self._name = child_cls.form_name
        assert("." not in self._name)
        self._tax_year = child_cls.tax_year
        self._inputs = inputs
        self._required_fields = required_fields
        self._optional_fields = optional_fields
        self._pdf_fields = pdf_fields
        self._pdf_file = pdf_file
        self._instance = instance
        self._solver = solver

        for i in self._inputs:
            i.__form_init__(self)
        for f in self._required_fields + self._optional_fields:
            f.__form_init__(self)

    def name(self):
        if self._instance is None:
            return self._name
        else:
            return f'{self._name}:{self._instance}'

    def full_description(self):
        return f'{self.__class__.description}: {self.__class__.long_description}'

    def instance(self):
        return self._instance

    def solver(self):
        assert(self._solver)
        return self._solver

    def inputs(self):
        return self._inputs

    def required_fields(self):
        return self._required_fields

    def fields(self):
        return self.required_fields() + self._optional_fields

    def pdf_fields(self):
        return self._pdf_fields

    def pdf_file(self):
        return self._pdf_file

    def needs_filing(self, values):
        raise NotImplementedError()

class InputForm(Form):
    """Convenience class to create a form which creates fields for each
    input"""
    def __init__(self,
                 child_cls,
                 inputs,
                 **kwargs):

        fields = []
        for i in inputs:
            base_name = i.base_name()
            fn = lambda s, i, v, base_name=base_name: i[base_name]

            if type(i) in [StringInput, SSNInput]:
                fields.append(StringField(base_name, fn))
            elif type(i) is BooleanInput:
                fields.append(BooleanField(base_name, fn))
            elif type(i) is IntegerInput:
                fields.append(IntegerField(base_name, fn))
            elif type(i) is FloatInput:
                fields.append(FloatField(base_name, fn))
            elif type(i) is EnumInput:
                fields.append(EnumField(base_name, i.enum, fn))
            else:
                raise TypeError(f'Unexpected input type in InputForm: {type(i)}')

        super().__init__(child_cls, inputs, fields, [], **kwargs)

    def needs_filing(self, values):
        # For typical input-only forms like W-2, 1099, 1098, habutax does not
        # need to present them to be filed because the user already has copies
        # of them
        return False

class FormAccessor(Mapping):
    def __init__(self, mapping, form):
        self.form = form
        self.mapping = mapping

    def __getitem__(self, key):
        if "." not in key:
            key = f'{self.form.name()}.{key}'
        return self.mapping[key]

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

def name_and_instance(full_form_name):
    split_form_name = full_form_name.split(':')
    form_instance = None
    if len(split_form_name) == 2:
        form_instance = split_form_name[1]
    elif len(split_form_name) != 1:
        raise RuntimeError(f'Unexpected form name: {full_form_name} (expected 0 or 1 colons)')
    return split_form_name[0], form_instance



from math import ceil
import os

from habutax.enum import filing_status_2021 as filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form1040S8812(Form):
    form_name = "1040_s8812"
    tax_year = 2021
    description = "Schedule 8812 (Form 1040)"
    long_description = "Credits for Qualifying Children and Other Dependents"
    jurisdiction = Jurisdiction.US
    sequence_no = 47

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('pr_income_forms_2555_4563', description="Did you exclude any income from Puerto Rico, or require Forms 2555 or 2563?"),
            IntegerInput('number_under_18', description="How many of your dependents eligible for the child tax credit with the required social security number were under age 18 at the end of 2021?"),
            IntegerInput('number_under_6', description="How many of your dependents eligible for the child tax credit with the required social security number were under age 6 at the end of 2021?"),
            BooleanInput('principal_abode_us', description="Did you (or your spouse if married filing jointly) have a principal place of abode in the United States for more than half of 2021?"),
            BooleanInput('resident_puerto_rico', description="Were you (or your spouse of married filing jointly) a bona fide resident of Puerto Rico for 2021?"),
            FloatInput('advance_ctc_payments', description="Enter the aggregate amount of advance child tax credit payments you (and your spouse if filing jointly) received for 2021. See your Letter(s) 6419 for the amounts to include)"),
            BooleanInput('credit_limit_ws_credits', description='Are you using Schedule 3 to report any nonrefundable credits, Forms 5695, 8910, 8936, or Schedule R?'),
            IntegerInput('number_children_letter', description="Enter the number of qualifying children taken into account in determining the annual advance child tax credit amount you received for 2021. See your Letter 6419 for this number. If you are missing your Letter 6419, you are filing a joint return, or you received more than one Letter 6419, see the instructions before entering a number"),
        ]

        def line_5_ws_6(self, i, v):
            if i['1040.filing_status'] is filing_status.MarriedFilingJointly:
                return 12500.0
            elif i['1040.filing_status'] is filing_status.QualifyingWidowWidower:
                return 2500.0
            elif i['1040.filing_status'] is filing_status.HeadOfHousehold:
                return 4375.0
            else:
                return 6250.0

        def line_5_ws_8(self, i, v):
            if i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingWidowWidower]:
                return 150000.0
            elif i['1040.filing_status'] is filing_status.HeadOfHousehold:
                return 112500.0
            else:
                return 75000.0

        def line_5_ws_9(self, i, v):
            res = v['3'] - v['5_ws_8']
            if res <= 0.001:
                return 0.0
            return ceil(res / 1000.0) * 1000.0

        def line_10(self, i, v):
            res = v['3'] - v['9']
            if res <= 0.001:
                return 0.0
            return ceil(res / 1000.0) * 1000.0

        def line_33(self, i, v):
            if i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingWidowWidower]:
                return 60000.0
            elif i['1040.filing_status'] is filing_status.HeadOfHousehold:
                return 50000.0
            else:
                return 40000.0

        def line_40(self, i, v):
            if v['32'] < 0.001 and v['32'] > -0.001:
                return v['29']
            return max(0.0, v['29'] - v['39'])

        def additional_tax(self, i, v):
            if v['14g'] > 0.001 or v['14g'] < -0.001:
                return 0.0
            if v['29'] < 0.001 and v['29'] > -0.001:
                return 0.0
            return v['40']

        optional_fields = [
            # Part I-A: Child Tax Credit and Credit for Other Dependents
            FloatField('1', lambda s, i, v: v['1040.11']),
            FloatField('2a', lambda s, i, v: s.not_implemented() if i['pr_income_forms_2555_4563'] else None),
            FloatField('2b', lambda s, i, v: s.not_implemented() if i['pr_income_forms_2555_4563'] else None),
            FloatField('2c', lambda s, i, v: s.not_implemented() if i['pr_income_forms_2555_4563'] else None),
            FloatField('2d', lambda s, i, v: v['2a'] + v['2b'] + v['2c']),
            FloatField('3', lambda s, i, v: v['1'] + v['2d']),
            IntegerField('4a', lambda s, i, v: i['number_under_18']),
            IntegerField('4b', lambda s, i, v: i['number_under_6']),
            IntegerField('4c', lambda s, i, v: v['4a'] - v['4b']),
            FloatField('5', lambda s, i, v: v['5_ws_12'] if v['4a'] > 0 else 0.0),
            IntegerField('6', lambda s, i, v: sum([v[f'1040.dependent_{n}_odc'] for n in range(i['1040.number_dependents'])])),
            FloatField('7', lambda s, i, v: v['6'] * 500.0),
            FloatField('8', lambda s, i, v: v['5'] + v['7']),
            FloatField('9', lambda s, i, v: 400000.0 if i['1040.filing_status'] is filing_status.MarriedFilingJointly else 200000.0),
            FloatField('10', line_10),
            FloatField('11', lambda s, i, v: v['10'] * 0.05),
            FloatField('12', lambda s, i, v: max(0.0, v['8'] - v['11'])),
            BooleanField('13a', lambda s, i, v: i['principal_abode_us']),
            BooleanField('13b', lambda s, i, v: i['resident_puerto_rico']),

            # Part I-B: Filers Who Check a Box on Line 13
            FloatField('14a', lambda s, i, v: min(v['7'], v['12'])),
            FloatField('14b', lambda s, i, v: v['12'] -  v['14a']),
            FloatField('14c', lambda s, i, v: 0.0 if v['14a'] < 0.001 else v['clwkst_a_5']),
            FloatField('14d', lambda s, i, v: min(v['14a'], v['14c'])),
            FloatField('14e', lambda s, i, v: v['14b'] + v['14d']),
            FloatField('14f', lambda s, i, v: i['advance_ctc_payments']),
            FloatField('14g', lambda s, i, v: max(0.0, v['14e'] - v['14f'])),
            FloatField('14h', lambda s, i, v: 0.0 if v['14g'] < 0.001 else min(v['14d'], v['14g'])),
            FloatField('14i', lambda s, i, v: 0.0 if v['14g'] < 0.001 else v['14g'] - v['14h']),

            FloatField('nonrefundable_ctc_or_odc', lambda s, i, v: v['14h'] if v['13a'] or v['13b'] else s.not_implemented()), # -> Form 1040, line 19
            FloatField('refundable_ctc_or_additional_ctc', lambda s, i, v: v['14i'] if v['13a'] or v['13b'] else s.not_implemented()), # -> Form 1040, line 28

            # Omitting Parts I-C, II-A, II-B, II-C - they are only used if you did *not* check a box on line 13
            FloatField('28a', lambda s, i, v: v['14f'] if v['13a'] or v['13b'] else v['15e']),
            FloatField('28b', lambda s, i, v: v['14e'] if v['13a'] or v['13b'] else v['15d']),
            FloatField('29', lambda s, i, v: v['28a'] - v['28b']),

            IntegerField('30', lambda s, i, v: i['number_children_letter']),
            IntegerField('31', lambda s, i, v: min(v['4a'], v['30'])),
            IntegerField('32', lambda s, i, v: v['30'] - v['31']),
            FloatField('33', line_33),
            FloatField('34', lambda s, i, v: max(0.0, v['3'] - v['33'])),
            FloatField('35', lambda s, i, v: v['33']),
            FloatField('36', lambda s, i, v: ceil((v['34'] / v['35']) / 1000.0) * 1000.0),
            FloatField('37', lambda s, i, v: v['32'] * 2000),
            FloatField('38', lambda s, i, v: v['37'] * v['36']),
            FloatField('39', lambda s, i, v: v['37'] - v['38']),
            FloatField('40', line_40),
            FloatField('additional_tax', additional_tax), # -> Schedule 2 (Form 1040), line 19
        ]

        optional_fields += [
            # Line 5 Worksheet from Schedule 8812 instructions
            FloatField('5_ws_1', lambda s, i, v: v['4b'] * 3600.0),
            FloatField('5_ws_2', lambda s, i, v: v['4c'] * 3000.0),
            FloatField('5_ws_3', lambda s, i, v: v['5_ws_1'] + v['5_ws_2']),
            FloatField('5_ws_4', lambda s, i, v: v['4a'] * 2000.0),
            FloatField('5_ws_5', lambda s, i, v: v['5_ws_3'] - v['5_ws_4']),
            FloatField('5_ws_6', line_5_ws_6),
            FloatField('5_ws_7', lambda s, i, v: min(v['5_ws_5'], v['5_ws_6'])),
            FloatField('5_ws_8', line_5_ws_8),
            FloatField('5_ws_9', line_5_ws_9),
            FloatField('5_ws_10', lambda s, i, v: v['5_ws_9'] * 0.05),
            FloatField('5_ws_11', lambda s, i, v: min(v['5_ws_7'], v['5_ws_10'])),
            FloatField('5_ws_12', lambda s, i, v: v['5_ws_3'] - v['5_ws_11']),

            # Credit Limit Worksheet A
            FloatField('clwkst_a_1', lambda s, i, v: v['1040.18']),
            FloatField('clwkst_a_2', lambda s, i, v: s.not_implemented() if i['credit_limit_ws_credits'] else None),
            FloatField('clwkst_a_3', lambda s, i, v: s.not_implemented() if not v['13a'] and not v['13b'] else v['clwkst_a_1'] - v['clwkst_a_2']),
            FloatField('clwkst_a_4', lambda s, i, v: s.not_implemented() if not v['13a'] and not v['13b'] else 0.0),
            FloatField('clwkst_a_5', lambda s, i, v: v['clwkst_a_3'] - v['clwkst_a_4']),
        ]

        required_fields = [
            BooleanField('dependents_match', lambda s, i, v: True if sum([v[f'1040.dependent_{n}_ctc'] for n in range(i['1040.number_dependents'])]) == v['4a'] else s.not_implemented()), # TODO would be more user-friendly to be reported as some type of error rather than "not implemented"
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_01[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_02[0]', '1040.you_ssn', max_length=11),
            TextPDFField('topmostSubform[0].Page1[0].f1_03[0]', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_04[0]', '2a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_05[0]', '2b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_06[0]', '2c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_07[0]', '2d'),
            TextPDFField('topmostSubform[0].Page1[0].f1_08[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_09[0]', '4a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '4b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_11[0]', '4c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '8'),
            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '10'),
            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '12'),
            ButtonPDFField('topmostSubform[0].Page1[0].Line13_ReadOrder[0].c1_1[0]', '13a', '1'),
            ButtonPDFField('topmostSubform[0].Page1[0].Line13_ReadOrder[0].c1_2[0]', '13b', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_20[0]', '14a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_21[0]', '14b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '14c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '14d'),
            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '14e'),
            TextPDFField('topmostSubform[0].Page1[0].f1_25[0]', '14f'),
            TextPDFField('topmostSubform[0].Page1[0].f1_26[0]', '14g'),
            TextPDFField('topmostSubform[0].Page1[0].f1_27[0]', '14h'),
            TextPDFField('topmostSubform[0].Page1[0].f1_28[0]', '14i'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_01[0]', '15a'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_02[0]', '15b'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_03[0]', '15c'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_04[0]', '15d'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_05[0]', '15e'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_06[0]', '15f'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_07[0]', '15g'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_08[0]', '15h'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_09[0]', '16a'),
#            TextPDFField('topmostSubform[0].Page2[0].Line16b_ReadOrder[0].f2_10[0]', '16b_num_qualifying_children'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_11[0]', '16b'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_12[0]', '17'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_13[0]', '18a'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_14[0]', '18b'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_1[0]', '19_no', '1'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_1[1]', '19_yes', '2'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_15[0]', '19'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_16[0]', '20'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_2[0]', '20_no', '1'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_2[1]', '20_yes', '2'),
#            TextPDFField('topmostSubform[0].Page2[0].Line21_ReadOrder[0].f2_17[0]', '21'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_18[0]', '22'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_19[0]', '23'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_20[0]', '24'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_21[0]', '25'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_22[0]', '26'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_23[0]', '27'),
            TextPDFField('topmostSubform[0].Page3[0].f3_01[0]', '28a'),
            TextPDFField('topmostSubform[0].Page3[0].f3_02[0]', '28b'),
            TextPDFField('topmostSubform[0].Page3[0].f3_03[0]', '29'),
            TextPDFField('topmostSubform[0].Page3[0].f3_04[0]', '30'),
            TextPDFField('topmostSubform[0].Page3[0].f3_05[0]', '31'),
            TextPDFField('topmostSubform[0].Page3[0].f3_06[0]', '32'),
            TextPDFField('topmostSubform[0].Page3[0].f3_07[0]', '33'),
            TextPDFField('topmostSubform[0].Page3[0].f3_08[0]', '34'),
            TextPDFField('topmostSubform[0].Page3[0].f3_09[0]', '35'),
            TextPDFField('topmostSubform[0].Page3[0].f3_10[0]', '36'),
            TextPDFField('topmostSubform[0].Page3[0].f3_11[0]', '37'),
            TextPDFField('topmostSubform[0].Page3[0].f3_12[0]', '38'),
            TextPDFField('topmostSubform[0].Page3[0].f3_13[0]', '39'),
            TextPDFField('topmostSubform[0].Page3[0].f3_14[0]', '40'),
        ]

        pdf_file = os.path.join(os.path.dirname(__file__), 'f1040s8.pdf')
        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



from .f1040 import Form1040
from .f1040_qualdiv_capgain_tax_wkst import Form1040QualDivCapGainTaxWkst
from .f1040_recovery_rebate_credit_wkst import Form1040RecoveryRebateCreditWkst
from .f1040_s2_need6251 import Form1040S2Need6251
from .f1040_s1 import Form1040S1
from .f1040_s3 import Form1040S3
from .f1040_sa import Form1040SA
from .f1040_sb import Form1040SB
from .f1040_s8812 import Form1040S8812
from .f1098 import Form1098
from .f1099_div import Form1099DIV
from .f1099_g import Form1099G
from .f1099_int import Form1099INT
from .f1099_r import Form1099R
from .f8606 import Form8606
from .f8959 import Form8959
from .f8995 import Form8995
from .fw_2 import FormW2
from .f8889 import Form8889
from .fnc_d_400 import FormNCD400
from .fnc_d_400_child_deduction_wkst import FormNCD400ChildDeductionWkst
from .fnc_d_400_consumer_use_tax_wkst import FormNCD400ConsumerUseTaxWkst
from .fnc_d_400_ss import FormNCD400SS
from .fnc_d_400_sa import FormNCD400SA

available_forms = [
	Form1040,
	Form1040QualDivCapGainTaxWkst,
	Form1040RecoveryRebateCreditWkst,
	Form1040S1,
	Form1040S2Need6251,
	Form1040S3,
	Form1040SA,
	Form1040SB,
	Form1040S8812,
	Form1098,
	Form1099DIV,
	Form1099G,
	Form1099INT,
	Form1099R,
	Form8606,
	Form8959,
	Form8995,
	FormW2,
	Form8889,
	FormNCD400,
	FormNCD400ChildDeductionWkst,
	FormNCD400ConsumerUseTaxWkst,
	FormNCD400SS,
	FormNCD400SA,
]



import os

from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form1040SB(Form):
    form_name = "1040_sb"
    tax_year = 2022
    description = "Schedule B (Form 1040)"
    long_description = "Interest and Ordinary Dividends"
    jurisdiction = Jurisdiction.US
    sequence_no = 8

    def __init__(self, **kwargs):
        NUM_FIELDS = 14
        inputs = [
            BooleanInput('excludable_us_bond_interest', description="Do you have any excludable interest on series EE and I U.S. savings bonds issued after 1989 to report?"),
            BooleanInput('financial_interest', description="At any time during 2022, did you have a financial interest in or signature authority over a financial account (such as a bank account, securities account, or brokerage account) located in a foreign country? See instructions for Form 1040, Schedule B."),
            BooleanInput('foreign_trust', description="During 2022, did you receive a distribution from, or were you the grantor of, or transferor to, a foreign trust? See instructions for Form 1040, Schedule B."),
        ]

        optional_fields = [
            FloatField('2', lambda s, i, v: sum([v[f'1_amount_{line}'] for line in range(NUM_FIELDS)])),
            FloatField('3', lambda s, i, v: s.not_implemented() if i['excludable_us_bond_interest'] else None),
            FloatField('4', lambda s, i, v: v['2'] - v['3']),
            FloatField('6', lambda s, i, v: sum([v[f'5_amount_{line}'] for line in range(NUM_FIELDS)])),
            BooleanField('7a', lambda s, i, v: s.not_implemented() if i['financial_interest'] else False),
            StringField('7b_country', lambda s, i, v: s.not_implemented() if i['financial_interest'] else None),
            BooleanField('7b', lambda s, i, v: s.not_implemented() if i['financial_interest'] else False),
            BooleanField('8', lambda s, i, v: s.not_implemented() if i['foreign_trust'] else False),
        ]

        # Because the other fields are pulled in as dependencies by Form 1040
        # when needed, and we want to allow part I and part II to be required
        # independently, add a check as a 'required field' (that will never be
        # filed) for part III, so that it is filled out when required.
        def part_3(self, i, v):
            if v['4'] > 1500.0 or v['6'] > 1500.0:
                v['7a']
                v['7b']
                v['8']
            # Also double-check that we don't have too many forms 1099-int or
            # 1099-div that we're trying to report here
            if i['1040.number_1099-int'] > NUM_FIELDS or i['1040.number_1099-div'] > NUM_FIELDS:
                self.not_implemented()
            return None

        required_fields = [
            StringField('part_3', part_3),
        ]

        for line in range(NUM_FIELDS):
            int_payer = StringField(f'1_payer_{line}', lambda s, i, v: v[f'1099-int:{s.which_1099int}.payer'] if s.which_1099int < i['1040.number_1099-int'] else None)
            int_payer.which_1099int = line
            int_amount = FloatField(f'1_amount_{line}', lambda s, i, v: v[f'1099-int:{s.which_1099int}.box_1'] + v[f'1099-int:{s.which_1099int}.box_3'] if s.which_1099int < i['1040.number_1099-int'] else None)
            int_amount.which_1099int = line
            div_payer = StringField(f'5_payer_{line}', lambda s, i, v: v[f'1099-div:{s.which_1099div}.payer'] if s.which_1099div < i['1040.number_1099-div'] else None)
            div_payer.which_1099div = line
            div_amount = FloatField(f'5_amount_{line}', lambda s, i, v: v[f'1099-div:{s.which_1099div}.box_1a'] if s.which_1099div < i['1040.number_1099-div'] else None)
            div_amount.which_1099div = line

            required_fields += [int_payer, int_amount, div_payer, div_amount]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_01[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_02[0]', '1040.you_ssn', max_length=11),
            TextPDFField('topmostSubform[0].Page1[0].Line1_ReadOrder[0].f1_03[0]', '1_payer_0'),
            TextPDFField('topmostSubform[0].Page1[0].f1_04[0]', '1_amount_0'),
            # Remaining input fields for field 1 are below in a 'for' loop
            TextPDFField('topmostSubform[0].Page1[0].f1_31[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_32[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_33[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].ReadOrderControl[0].f1_34[0]', '5_payer_0'),
            TextPDFField('topmostSubform[0].Page1[0].f1_35[0]', '5_amount_0'),
            # Remaining input fields for field 5 are below in a 'for' loop
            TextPDFField('topmostSubform[0].Page1[0].f1_64[0]', '6'),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[0]', '7a', '1'),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[1]', '7a', '2', lambda s, v, f: not v),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_2[0]', '7b', '1'),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_2[1]', '7b', '2', lambda s, v, f: not v),
            TextPDFField('topmostSubform[0].Page1[0].f1_65[0]', '7b_country'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_66[0]', '7b_country'),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_3[0]', '8', '1'),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_3[1]', '8', '2', lambda s, v, f: not v),
        ]

        # The first of each of the fields in this loop is handled above,
        # because they require different "selectors"
        for line in range(1, NUM_FIELDS):
            one_payer = TextPDFField(f'topmostSubform[0].Page1[0].f1_{3+line*2:02d}[0]', f'1_payer_{line}')
            one_amt = TextPDFField(f'topmostSubform[0].Page1[0].f1_{4+line*2:02d}[0]', f'1_amount_{line}')
            five_payer = TextPDFField(f'topmostSubform[0].Page1[0].f1_{34+line*2:02d}[0]', f'5_payer_{line}')
            five_amt = TextPDFField(f'topmostSubform[0].Page1[0].f1_{35+line*2:02d}[0]', f'5_amount_{line}')
            pdf_fields += [one_payer, one_amt, five_payer, five_amt]

        pdf_file = os.path.join(os.path.dirname(__file__), 'f1040sb.pdf')
        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



import os

from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form8606(Form):
    form_name = "8606"
    tax_year = 2022
    description = "Form 8606"
    long_description = "Nondeductible IRAs"
    jurisdiction = Jurisdiction.US
    sequence_no = 48
    valid_instances = ['you', 'spouse']

    def __init__(self, **kwargs):
        instance = kwargs['instance']
        assert(instance in ['you', 'spouse'])
        you = "you" if instance == "you" else "your spouse"
        your = "your" if instance == "you" else "your spouse's"

        inputs = [
            BooleanInput('part_1_needed', description=f'Do {you} need to fill out part 1 of Form 8606? This is required if one or more of the following apply: 1) {you} made nondeductible contributions to a traditional IRA for 2022, 2) {you} took distributions from a traditional, SEP, or SIMPLE IRA in 2022 and {you} made nondeductible contributions to a traditional IRA in 2022 or an earlier year, 3) {you} converted part, but not all, of {your} traditional, SEP, and SIMPLE IRAs to Roth IRAs in 2022 and {you} made nondeductible contributions to a traditional IRA in 2022 or an earlier year. See Form 8606 instructions for more information.'),
            FloatInput('nondeductible_contributions', description=f'Enter {your} nondeductible contributions to traditional IRAs for 2022, including those made for 2022 from January 1, 2023, through April 15, 2023'),
            FloatInput('traditional_basis', description=f'Enter {your} total basis in traditional IRAs. See instructions'),
            BooleanInput('distribution_or_roth_conversion', description=f'In {Form8606.tax_year}, did {you} take a distribution from traditional, SEP, or SIMPLE IRAs, or make a Roth IRA conversion?'),
            FloatInput('nondeductible_contributions_next_year', description=f'Enter {your} nondeductible contributions to traditional IRAs for 2022 made from January 1, 2023 through April 15, 2023'),
            FloatInput('year_end_value_non_roth', description=f'Enter the value of *all* {your} traditional, SEP, and SIMPLE IRAs as of December 31, 2022, plus any outstanding rollovers. Subtract any repayments of qualified disaster distributions (see 2022 Forms 8915-D and 8915-F)'),
            FloatInput('distributions_2022', description=f'Enter {your} distributions from traditional, SEP, and SIMPLE IRAs in 2022. Do *not* include rollovers (other than repayments of qualified disaster distributions (see 2022 Forms 8915-D and 8915-F)), qualified charitable distributions, a one-time distribution to fund an HSA, conversions to a Roth IRA, certain returned contributions, or recharacterizations of traditional IRA contributions (see instructions)'),
            BooleanInput('qualified_disaster_distributions', description=f'Are any of {your} distributions from traditional, SEP, and Simple IRAs in 2022 qualified disaster distributions?'),
            FloatInput('net_converted', description=f'Enter the net amount {you} converted from traditional, SEP, and SIMPLE IRAs to Roth IRAs in 2022.'),
            BooleanInput('part_2_needed', description=f'Did {you} convert part or all of {your} traditional, SEP, and SIMPLE IRAs to a Roth IRA in {Form8606.tax_year}?'),
            FloatInput('converted_cost_basis', description=f'Enter {your} cost basis for the net amount {you} converted from traditional, SEP, and SIMPLE IRAs to Roth IRAs in {Form8606.tax_year}'),
            BooleanInput('part_3_needed', description=f'Did {you} take a distribution from a Roth IRA in 2022?'),
            FloatInput('total_nonqualified_distributions', description=f'Enter {your} total nonqualified distributions from Roth IRAs in 2022, including any qualified first-time homebuyer distributions, and any qualified disaster distributions (see instructions). Also, see 2022 Forms 8915-D and 8915-F'),
            FloatInput('qualified_homebuyer', description=f'Qualified first-time homebuyer expenses (see instructions). Do not enter more than $10,000 reduced by the total of all {your} prior qualified first-time homebuyer distributions'),
            FloatInput('roth_ira_contributions_basis', description=f'Enter {your} basis in Roth IRA contributions (see instructions)'),
        ]

        # All fields are optional. Form 1040/1040-SR/1040-NR should include the
        # sum of 'taxable_amount' fields from all forms 8606 on line 4b.
        optional_fields = [
            FloatField('1', lambda s, i, v: i['nondeductible_contributions']),
            FloatField('2', lambda s, i, v: i['traditional_basis']),
            FloatField('3', lambda s, i, v: v['1'] + v['2']),
            FloatField('4', lambda s, i, v: i['nondeductible_contributions_next_year']),
            FloatField('5', lambda s, i, v: v['3'] - v['4']),
            FloatField('6', lambda s, i, v: i['year_end_value_non_roth']),
            FloatField('7', lambda s, i, v: i['distributions_2022']),
            FloatField('8', lambda s, i, v: i['net_converted']),
            FloatField('9', lambda s, i, v: v['6'] + v['7'] + v['8']),
            FloatField('10', lambda s, i, v: min(1.000, float(v['5']) / float(v['9']), 1.000), places=5),
            FloatField('11', lambda s, i, v: v['8'] * v['10']),
            FloatField('12', lambda s, i, v: v['7'] * v['10']),
            FloatField('13', lambda s, i, v: v['11'] + v['12']),
            FloatField('14', lambda s, i, v: v['3'] - v['13'] if i['distribution_or_roth_conversion'] else v['3']),
            FloatField('15a', lambda s, i, v: v['7'] - v['12']),
            FloatField('15b', lambda s, i, v: self.not_implemented() if i['qualified_disaster_distributions'] else None),
            FloatField('15c', lambda s, i, v: v['15a'] - v['15b']),
            FloatField('16', lambda s, i, v: i['net_converted']),
            FloatField('17', lambda s, i, v: v['11'] if i['part_1_needed'] else i['converted_cost_basis']),
            FloatField('18', lambda s, i, v: v['16'] - v['17']),
            FloatField('19', lambda s, i, v: i['total_nonqualified_distributions']),
            FloatField('20', lambda s, i, v: i['qualified_homebuyer']),
            FloatField('21', lambda s, i, v: max(0.0, v['19'] - v['20'])),
            FloatField('22', lambda s, i, v: i['roth_ira_contributions_basis']),
            FloatField('23', lambda s, i, v: max(0.0, v['21'] - v['22'])),
        ]

        # Check the (part_1_needed, part_2_needed, part_3_needed inputs
        # from this form and pull the needed values (lines 15c, 18, and/or
        # 25c) directly to cause those sections to be calculated
        def taxable_amount(self, i, v):
            total = 0.0
            if i['part_1_needed']:
                if i['distribution_or_roth_conversion']:
                    total += v['15c']
                v['14'] # Not returned, but needs to be populated
            if i['part_2_needed']:
                total += v['18']
            if i['part_3_needed']:
                if v['21'] < 0.001:
                    v['22'] # Not returned, but needs to be populated
                elif v['23'] >= 0.001:
                    # Generally, there is an additional 10% tax on 2022
                    # distributions from a Roth IRA that are shown on line 23.
                    # You will need to complete lines 1 through 4 of Form 5329
                    # to determine the amounts from the Roth IRAs that are
                    # subject to the additional tax. See the instructions for
                    # Form 5329, Part I, for details and exceptions.
                    self.not_implemented()
            return total

        def name_field(self, i, v):
            if self.form().instance() == 'spouse':
                return f'{v["1040.spouse_first_name"]} {v["1040.spouse_last_name"]}'
            else:
                return f'{v["1040.first_name"]} {v["1040.last_name"]}'

        required_fields = [
            StringField('name', name_field),
            StringField('ssn', lambda s, i, v: v[f'1040.{s.form().instance()}_ssn']),
            FloatField('taxable_amount', taxable_amount),
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_1[0]', 'name'),
            TextPDFField('topmostSubform[0].Page1[0].f1_2[0]', 'ssn', max_length=11),
            # only if filling form by itself
#            TextPDFField('topmostSubform[0].Page1[0].f1_3[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_4[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_5[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_6[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_7[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_8[0]', 'unknown'),
            TextPDFField('topmostSubform[0].Page1[0].f1_9[0]', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_11[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '8'),
            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '10', max_length=1, value_fn=lambda s, v, f: f'{v:.3f}'.split('.')[0]),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '10', value_fn=lambda s, v, f: f'{v:.3f}'.split('.')[1]),
            TextPDFField('topmostSubform[0].Page1[0].f1_20[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_21[0]', '12'),
            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '13'),
            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '14'),
            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '15a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_25[0]', '15b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_26[0]', '15c'),
            TextPDFField('topmostSubform[0].Page2[0].f2_1[0]', '16'),
            TextPDFField('topmostSubform[0].Page2[0].f2_2[0]', '17'),
            TextPDFField('topmostSubform[0].Page2[0].f2_3[0]', '18'),
            TextPDFField('topmostSubform[0].Page2[0].f2_4[0]', '19'),
            TextPDFField('topmostSubform[0].Page2[0].f2_5[0]', '20'),
            TextPDFField('topmostSubform[0].Page2[0].f2_6[0]', '21'),
            TextPDFField('topmostSubform[0].Page2[0].f2_7[0]', '22'),
            TextPDFField('topmostSubform[0].Page2[0].f2_8[0]', '23'),
            # unimplemented
#            TextPDFField('topmostSubform[0].Page2[0].f2_9[0]', '24'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_10[0]', '25a'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_11[0]', '25b'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_12[0]', '25c'),
            # only if filling form by itself
#            TextPDFField('topmostSubform[0].Page2[0].f2_13[0]', 'unknown'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_1[0]', 'unknown', '1'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_14[0]', 'unknown', max_length=11),
#            TextPDFField('topmostSubform[0].Page2[0].f2_15[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_16[0]', 'unknown'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_17[0]', 'unknown', max_length=10),
#            TextPDFField('topmostSubform[0].Page2[0].f2_18[0]', 'unknown'),
        ]

        pdf_file = os.path.join(os.path.dirname(__file__), 'f8606.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



import unittest

from ..test_form import FormTestCase

import habutax.forms.ty2022 as ty2022

class Form1040S3TestCase(FormTestCase):
    def setUp(self):
        values = {
            '1040.11': 83450.73,
            '1099-int:0.box_6': 329.09,
            '1099-div:0.box_7': 108.14,
        }
        self.fixture_setup(ty2022.available_forms, input_fixture='tests/ty2022/fixtures/f1040_s3_0.habutax', values=values)

    def test_1040_s3(self):
        solution = self.assertSolve(['1040_s3'], field_names=['1040_s3.8'])
        self.assertIn('1040_s3', solution)

        self.assertDollarsEqual(solution, '1040_s3.1', 437.23)
        self.assertDollarsEqual(solution, '1040_s3.8', 437.23)

class Form1040S3TestCaseTooManyForeignTaxesMFJ(FormTestCase):
    def setUp(self):
        values = {
            '1040.11': 83450.73,
            '1099-int:0.box_6': 329.09,
            '1099-div:0.box_7': 423.09,
        }
        self.fixture_setup(ty2022.available_forms, input_fixture='tests/ty2022/fixtures/f1040_s3_0.habutax', values=values)

    def test_1040_s3(self):
        solution = self.assertUnimplementedSolution(['1040_s3'], field_names=['1040_s3.8'])

class Form1040S3TestCaseTooManyForeignTaxesSingle(FormTestCase):
    def setUp(self):
        inputs = {
            '1040.filing_status': 'Single'
        }
        values = {
            '1040.11': 83450.73,
            '1099-int:0.box_6': 329.09,
            '1099-div:0.box_7': 123.09,
        }
        self.fixture_setup(ty2022.available_forms, input_fixture='tests/ty2022/fixtures/f1040_s3_0.habutax', inputs=inputs, values=values)

    def test_1040_s3(self):
        solution = self.assertUnimplementedSolution(['1040_s3'], field_names=['1040_s3.8'])



import unittest

from ..test_form import FormTestCase

import habutax.forms.ty2021 as ty2021

class Form1040S1TestCase0(FormTestCase):
    def setUp(self):
        values = {
            '1099-g:0.box_2': 329.09,
            '1099-g:1.box_2': 108.14,
            '1098:0.box_4': 102.33,
            '1099-int:0.box_2': 53.98,
            '8889:spouse.hsa_deduction': 7200.00,
        }
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_s1_0.habutax', values=values)

    def test_1040_s1(self):
        solution = self.assertSolve(['1040_s1'], field_names=['1040_s1.10', '1040_s1.26'])

        self.assertIn('1040_s1', solution)

        self.assertDollarsEqual(solution, '1040_s1.1', 437.23)
        self.assertDollarsEqual(solution, '1040_s1.2a', 0.00)
        self.assertStringFieldEqual(solution, '1040_s1.2b', '')
        self.assertDollarsEqual(solution, '1040_s1.7', 0.00)
        self.assertDollarsEqual(solution, '1040_s1.8z', 102.33)
        self.assertStringFieldEqual(solution, '1040_s1.8z_type', "Refund of overpaid mortgage interest")
        self.assertDollarsEqual(solution, '1040_s1.10', 539.56)
        self.assertDollarsEqual(solution, '1040_s1.11', 0.00)
        self.assertDollarsEqual(solution, '1040_s1.13', 7200.00)
        self.assertDollarsEqual(solution, '1040_s1.18', 53.98)
        self.assertDollarsEqual(solution, '1040_s1.19a', 4983.2)
        self.assertStringFieldEqual(solution, '1040_s1.19b', '123456789')
        self.assertStringFieldEqual(solution, '1040_s1.19c', '04/01/2011')
        self.assertDollarsEqual(solution, '1040_s1.24z', 123.45)
        self.assertStringFieldEqual(solution, '1040_s1.24z_type', 'something')
        self.assertDollarsEqual(solution, '1040_s1.25', 123.45)
        self.assertDollarsEqual(solution, '1040_s1.26', 12360.63)

class Form1040S1TestCase1(FormTestCase):
    def setUp(self):
        values = {
        }
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_s1_1.habutax')

    def test_1040_s1(self):
        solution = self.assertSolve(['1040_s1'], field_names=['1040_s1.10', '1040_s1.26'])

        self.assertIn('1040_s1', solution)

        self.assertDollarsEqual(solution, '1040_s1.1', 0.00)
        self.assertDollarsEqual(solution, '1040_s1.2a', 1245.46)
        self.assertStringFieldEqual(solution, '1040_s1.2b', '12/10/2018')
        self.assertDollarsEqual(solution, '1040_s1.7', 943.00)
        self.assertDollarsEqual(solution, '1040_s1.10', 2188.46)
        self.assertDollarsEqual(solution, '1040_s1.11', 234.50)
        self.assertDollarsEqual(solution, '1040_s1.13', 0.0)
        self.assertDollarsEqual(solution, '1040_s1.24z', 0.0)
        self.assertStringFieldEqual(solution, '1040_s1.24z_type', '')
        self.assertDollarsEqual(solution, '1040_s1.25', 0.0)
        self.assertDollarsEqual(solution, '1040_s1.26', 234.50)

class Form1040S1TestCase1_no_hsa(FormTestCase):
    def setUp(self):
        inputs = {
            '1040.filing_status': 'MarriedFilingJointly',
            '1040_s1.hsa_contribution_you': 'no',
            '1040_s1.hsa_contribution_spouse': 'no',
        }
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_s1_1.habutax', inputs=inputs)

    def test_1040_s1(self):
        solution = self.assertSolve(['1040_s1'], field_names=['1040_s1.26'])
        self.assertDollarsEqual(solution, '1040_s1.13', 0.0)

class Form1040S1TestCase1_you_hsa(FormTestCase):
    def setUp(self):
        inputs = {
            '1040.filing_status': 'MarriedFilingJointly',
            '1040_s1.hsa_contribution_you': 'yes',
            '1040_s1.hsa_contribution_spouse': 'no',
        }
        values = {
            '8889:you.hsa_deduction': 198.30,
        }
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_s1_1.habutax', inputs=inputs, values=values)

    def test_1040_s1(self):
        solution = self.assertSolve(['1040_s1'], field_names=['1040_s1.26'])
        self.assertDollarsEqual(solution, '1040_s1.13', 198.30)

class Form1040S1TestCase1_spouse_hsa(FormTestCase):
    def setUp(self):
        inputs = {
            '1040.filing_status': 'MarriedFilingJointly',
            '1040_s1.hsa_contribution_you': 'no',
            '1040_s1.hsa_contribution_spouse': 'yes',
        }
        values = {
            '8889:spouse.hsa_deduction': 1250.00,
        }
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_s1_1.habutax', inputs=inputs, values=values)

    def test_1040_s1(self):
        solution = self.assertSolve(['1040_s1'], field_names=['1040_s1.26'])
        self.assertDollarsEqual(solution, '1040_s1.13', 1250.00)

class Form1040S1TestCase1_both_hsa(FormTestCase):
    def setUp(self):
        inputs = {
            '1040.filing_status': 'MarriedFilingJointly',
            '1040_s1.hsa_contribution_you': 'yes',
            '1040_s1.hsa_contribution_spouse': 'yes',
        }
        values = {
            '8889:spouse.hsa_deduction': 1250.00,
            '8889:you.hsa_deduction': 3892.00,
        }
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_s1_1.habutax', inputs=inputs, values=values)

    def test_1040_s1(self):
        solution = self.assertSolve(['1040_s1'], field_names=['1040_s1.26'])
        self.assertDollarsEqual(solution, '1040_s1.13', 5142.00)



import os

from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form8889(Form):
    form_name = "8889"
    tax_year = 2022
    description = "Form 8889"
    long_description = "Health Savings Accounts (HSAs)"
    jurisdiction = Jurisdiction.US
    sequence_no = 52
    valid_instances = ['you', 'spouse']

    def __init__(self, **kwargs):
        instance = kwargs['instance']
        assert(instance in ['you', 'spouse'])
        you = "you" if instance == "you" else "your spouse"
        your = "your" if instance == "you" else "your spouse's"
        your_spouses = "your spouse's" if instance == "you" else "your"

        inputs = [
            BooleanInput('part_2_needed', description=f'Do {you} need to fill out part 2 of Form 8999? Part 2 is used to report HSA distributions.'),
            BooleanInput('part_3_needed', description=f'Do {you} need to fill out part 3 of Form 8999? Part 3 is used to report cases when you fail to maintain high-deductible health plan (HDHP) coverage for the entire year.  See instructions.'),
            BooleanInput('hdhp_plan_family', description=f'Were you covered under a *family* high-deductible health plan (HDHP) during 2022 (as opposed to self-only coverage)? If {you} were covered, or considered covered, by a self-only high-deductible health plan (HDHP) and a family HDHP at different times during the year, answer "yes" only if the family plan was in effect for a longer period. If {you} were covered by both a self-only HDHP and a family HDHP at the same time, {you} are treated as having family coverage during that period. If, on the first day of the last month of {your} tax year (December 1 for most taxpayers), {you} had family coverage, answer "yes". Were you covered under a high-deductible *family* plan during 2022?'),
            FloatInput('hsa_contributions', description=f'Enter HSA contributions {you} made for 2022 (or those made on {your} behalf), including those made from January 1, 2022, through April 15, 2022, that were for 2022. Do not include employer contributions, contributions through a cafeteria plan, or rollovers. See instructions for Form 8889, line 2.'),
            BooleanInput('age_under_55', description=f'Were {you} under the age of 55 at the end of 2022?'),
            BooleanInput('hsa_full_year', description=f'Did {you} have a HDHP medical plan for the first day of every month during the 2022 year?'),
            FloatInput('archer_msa', description=f'Enter the amount {you} and {your} employer contributed to {your} Archer MSAs for 2022 from Form 8853, lines 1 and 2. If you or your spouse had family coverage under an HDHP at any time during 2022, also include any amount contributed to {your_spouses} Archer MSAs'),
            BooleanInput('qualified_distribution', description=f'Did {you} take a distribution from {your} traditional IRA or Roth IRA to {your} HSA in a direct trustee-to-trustee transfer?'),
            FloatInput('employer_contribution', description=f'Employer contributions made to {your} HSAs for 2022? NOTE: This number should match the sum of all {your} your forms W-2, box 12, code W.'),
        ]

        def calc_contribution_max(self, i, v):
            if i['age_under_55'] and i['hsa_full_year']:
                return 7300 if v['1'] else 3650
            else:
                self.not_implemented()

        # All fields are optional. Form 1040s1 13 should include both you and
        # your spouse hsa_deductions
        optional_fields = [
            BooleanField('1', lambda s, i, v: i['hdhp_plan_family']),
            FloatField('2', lambda s, i, v: i['hsa_contributions']),
            IntegerField('3', calc_contribution_max),
            FloatField('4', lambda s, i, v: s.not_implemented() if i['archer_msa'] else 0.0),
            FloatField('5', lambda s, i, v: max(v['3'] - v['4'], 0.0)),
            FloatField('6', lambda s, i, v: s.not_implemented() if i['1040_s1.hsa_contribution_you'] and i['1040_s1.hsa_contribution_spouse'] and i['hdhp_plan_family'] else v['5']),
            FloatField('7', lambda s, i, v: s.not_implemented() if not i['age_under_55'] and (i['hdhp_plan_family'] or i['8889:spouse.hdhp_plan_family']) else None),
            FloatField('8', lambda s, i, v: v['6'] + v['7']),
            FloatField('9', lambda s, i, v: i['employer_contribution']),
            FloatField('10', lambda s, i, v: s.not_implemented() if i['qualified_distribution'] else None),
            FloatField('11', lambda s, i, v: v['9'] + v['10']),
            FloatField('12', lambda s, i, v: max(v['8'] - v['11'], 0.0)),
            FloatField('13', lambda s, i, v: min(v['2'], v['12'])),
            FloatField('hsa_deduction', lambda s, i, v: s.not_implemented() if i['part_2_needed'] or i['part_3_needed'] or v['2'] > v['13'] else v['13']),
        ]

        def name_field(self, i, v):
            if self.form().instance() == 'spouse':
                return f'{v["1040.spouse_first_name"]} {v["1040.spouse_last_name"]}'
            else:
                return f'{v["1040.first_name"]} {v["1040.last_name"]}'

        required_fields = [
            StringField('name', name_field),
            StringField('ssn', lambda s, i, v: v[f'1040.{s.form().instance()}_ssn']),
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_01[0]', 'name'),
            TextPDFField('topmostSubform[0].Page1[0].f1_02[0]', 'ssn', max_length=11),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[0]', '1', '1', value_fn=lambda s, v, f: not v),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[1]', '1', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_03[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_04[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_05[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].f1_06[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].f1_07[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_08[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_09[0]', '8'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_11[0]', '10'),
            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '12'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '13'),
            # parts II, III unimplemented
#            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '14a'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '14b'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '14c'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '15'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '16'),
#            ButtonPDFField('topmostSubform[0].Page1[0].c1_2[0]', '17a', '1'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_20[0]', '17b'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_21[0]', '18'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '19'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '20'),
#            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '21'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'f8889.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



import os

from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form8995(Form):
    form_name = "8995"
    tax_year = 2021
    description = "Form 8995"
    long_description = "Qualified Business Income Deduction Simplified Computation"
    jurisdiction = Jurisdiction.US
    sequence_no = 55

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('other_than_199a', description='Do you have any qualified business income (other than section 199a dividends) you need to report for this year or carryover from a previous year you need to report for qualified busiess income of REIT dividends and PTP income, or any reason you cannot claim the 199a dividends reported on your forms 1099-DIV?'),
            BooleanInput('schedule_d', description='Are you required to file schedule D?'),
        ]

        optional_fields = [
            FloatField('2', lambda s, i, v: s.not_implemented() if i['other_than_199a'] else None),
            FloatField('3', lambda s, i, v: s.not_implemented() if i['other_than_199a'] else None),
            FloatField('4', lambda s, i, v: max(0.0, v['2'] + v['3'])),
            FloatField('5', lambda s, i, v: v['4'] * 0.20),
            FloatField('6', lambda s, i, v: sum([v[f'1099-div:{n}.box_5'] for n in range(i['1040.number_1099-div'])])),
            FloatField('7', lambda s, i, v: s.not_implemented() if i['other_than_199a'] else None),
            FloatField('8', lambda s, i, v: max(0.0, v['6'] + v['7'])),
            FloatField('9', lambda s, i, v: v['8'] * 0.20),
            FloatField('10', lambda s, i, v: v['5'] + v['9']),
            FloatField('11', lambda s, i, v: v['1040.11'] - v['1040.12c']),
            FloatField('12', lambda s, i, v: v['1040.3a'] + v['1040.7'] if not i['schedule_d'] else s.not_implemented()),
            FloatField('13', lambda s, i, v: max(0.0, v['11'] - v['12'])),
            FloatField('14', lambda s, i, v: v['13'] * 0.20),
            FloatField('15', lambda s, i, v: min(v['10'], v['14'])),
        ]

        required_fields = [
            FloatField('16', lambda s, i, v: min(0.0, v['2'] + v['3'])),
            FloatField('17', lambda s, i, v: min(0.0, v['6'] + v['7'])),
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_1[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_2[0]', '1040.you_ssn', max_length=11),
            # not implemented
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1A_Row1[0].f1_3[0]', '1a_i'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1A_Row1[0].f1_4[0]', '1b_i', max_length=11),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1A_Row1[0].f1_5[0]', '1c_i'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1B_Row2[0].f1_6[0]', '1a_ii'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1B_Row2[0].f1_7[0]', '1b_ii', max_length=11),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1B_Row2[0].f1_8[0]', '1c_ii'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1C_Row3[0].f1_9[0]', '1a_iii'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1C_Row3[0].f1_10[0]', '1b_iii', max_length=11),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1C_Row3[0].f1_11[0]', '1c_iii'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1D_Row4[0].f1_12[0]', '1a_iv'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1D_Row4[0].f1_13[0]', '1b_iv', max_length=11),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1D_Row4[0].f1_14[0]', '1c_iv'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1E_Row5[0].f1_15[0]', '1a_v'),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1E_Row5[0].f1_16[0]', '1b_v', max_length=11),
#            TextPDFField('topmostSubform[0].Page1[0].Table[0].Ln1E_Row5[0].f1_17[0]', '1c_v'),
            TextPDFField('topmostSubform[0].Page1[0].ReadOrderSubForm[0].f1_18[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_20[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].f1_21[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '8'),
            TextPDFField('topmostSubform[0].Page1[0].f1_25[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_26[0]', '10'),
            TextPDFField('topmostSubform[0].Page1[0].f1_27[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_28[0]', '12'),
            TextPDFField('topmostSubform[0].Page1[0].f1_29[0]', '13'),
            TextPDFField('topmostSubform[0].Page1[0].f1_30[0]', '14'),
            TextPDFField('topmostSubform[0].Page1[0].f1_31[0]', '15'),
            TextPDFField('topmostSubform[0].Page1[0].f1_32[0]', '16'),
            TextPDFField('topmostSubform[0].Page1[0].f1_33[0]', '17'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'f8995.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True
